-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

local _G = getfenv(0)

local deformat = AceLibrary("Deformat-2.0")
local tablet = AceLibrary("Tablet-2.0")
local roster = AceLibrary("Roster-2.1")
local BC = AceLibrary("Babble-Class-2.2")
local BB = AceLibrary("Babble-Boss-2.2")
local ctl = ChatThrottleLib

local running = nil
local db = nil

local coloredName = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		local class = roster.roster[key] and roster.roster[key].class or "WTF"
		self[key] = "|cff" .. BC:GetHexColor(class) .. key .. "|r"
		return self[key]
	end
})

local dkpValues = setmetatable({}, {__index =
	function(self, key)
		if type(key) == "nil" then return nil end
		self[key] = BananaDKP:GetDKP(key) or 0
		return self[key]
	end
})

local whisperHelp = nil
local whisperHandlers = nil
local commHandlers = {}
local commInitialized = nil

local colors = {
	["progress"] = "|cff69ccf0",
	["fail"] = "|cffff1919",
	["success"] = "|cff19ff19",
	["inactive"] = "|cff9d9d9d",
}

-- Success
local STATUS_ACTIVE = 1

-- Inactive
local STATUS_NO_STATUS = 100
local STATUS_NO_LEADER = 101
local STATUS_VALID_NETWORK = 102
local STATUS_INACTIVE = 103

-- Progress
local STATUS_SETTING_LEADER = 201
local STATUS_CHECKING_LEADER = 202
local STATUS_DKP_SYNC = 203
local STATUS_CHECKING_NETWORK = 204
local STATUS_TAKING_BIDS = 205

-- Failure
local STATUS_NO_DKP_FROM_LEADER = 301
local STATUS_INVALID_NETWORK = 302
local STATUS_NO_ID_SET = 303
local STATUS_VERSION_MISMATCH = 304
local STATUS_LEADER_DISPUTE = 305

local currentStatus = STATUS_NO_STATUS

-- Participant locals.
local acceptedLeader = nil
local currentLeader = nil

-- Leader locals.
local rejectAsLeader = 0
local recentWhispers = nil
local winner = nil

-- If we ever change communication implementation details or the database
-- implementation details, we have to modify these.
_G.BANANADKP_COMM_VERSION = 1

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("BananaDKP")
L:RegisterTranslations("enUS", function() return {
	[1] = "Monitoring",

	[100] = "Disabled",
	[101] = "No leader",
	[102] = "Valid network",
	[103] = "Inactive",

	[201] = "Setting leader ...",
	[202] = "Checking leader ...",
	[203] = "Synchronizing ...",
	[204] = "Checking network ...",
	[205] = "Taking bids ...",

	[301] = "No DKP from leader!",
	[302] = "Invalid network!",
	[303] = "No DKP ID set!",
	[304] = "Version mismatch!",
	[305] = "Leader dispute!",

	["Unknown status!"] = true,

	["Status"] = true,
	["DKP Leader"] = true,
	["You"] = true,
	["DKP Type"] = true,
	["Player"] = true,
	["DKP"] = true,

	["Stream identifier"] = true,
	["Set the DKP stream identifier to use for data synchronization. This must be the same for everyone."] = true,
	["Type"] = true,
	["Set what DKP system to use. Everyone must use the same system."] = true,
	["Type options"] = true,
	["Options for the current DKP type."] = true,
	["Enable BananaDKP"] = true,
	["Start up BananaDKP for this raid session."] = true,
	["Set myself as leader"] = true,
	["If there's no DKP stream leader in the raid, and we're an officer or raid leader, we can set ourselves as the DKP stream leader."] = true,
	["Synchronize DKP"] = true,
	["Synchronizes the DKP stream with your current data. You can only do this if you're the leader, and it is done automatically one time when you become the leader."] = true,
	["Accept"] = true,
	["Reject"] = true,
	["%s wants to be the DKP stream leader for this raid.\n\nDo you accept? (You will auto-accept in 20 seconds.)"] = true,
	["Activate %s."] = true,
	["Sitouts"] = true,
	["List of players sitting out this raid who should still receive DKP. Click on a name to remove him from the list."] = true,
	["Add player"] = true,
	["Adds a player to the sitout list."] = true,
	["Clear list"] = true,
	["Clear the sitout list."] = true,
	["Remove %s."] = true,
	["<player name>"] = true,

	["BananaDKP: %s."] = true,
	["BananaDKP: DKP for %s: %s."] = true,
	["BananaDKP: Unknown command or player not found."] = true,
	["^bid:"] = true,
	["^bid"] = true,

	["%s wins!"] = true,
	["Tie: %s."] = true,

	["Chat-Help-Dkp"] = "Whisper me 'bdkp (<player>)' to get the current DKP for the given player or you.",
	["Chat-Help-Bid"] = "When the raid leader says 'bid: <item>' in raid chat, you can just say 'bid' in whisper to him or raid chat in order to register yourself as a bidder.",
} end)

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local options = {
	type = "group",
	args = {
		dkpType = {
			type = "group",
			name = L["Type"],
			desc = L["Set what DKP system to use. Everyone must use the same system."],
			args = {},
			order = 101,
			disabled = function() return running end,
		},
		dkpTypeOptions = {
			type = "group",
			name = L["Type options"],
			desc = L["Options for the current DKP type."],
			args = {},
			order = 102,
			disabled = function()
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				if type(currentLeader) ~= "boolean" or currentLeader ~= true then return true end
				return currentStatus == STATUS_DKP_SYNC
			end,
		},
		spacer = {
			type = "header",
			name = " ",
			order = 103,
		},
		setleader = {
			type = "execute",
			name = L["Set myself as leader"],
			desc = L["If there's no DKP stream leader in the raid, and we're an officer or raid leader, we can set ourselves as the DKP stream leader."],
			func = "SetSelfAsLeader",
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				return currentStatus ~= STATUS_NO_LEADER
			end,
			order = 104,
		},
		sync = {
			type = "execute",
			name = L["Synchronize DKP"],
			desc = L["Synchronizes the DKP stream with your current data. You can only do this if you're the leader, and it is done automatically one time when you become the leader."],
			func = "SyncRaidDKP",
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				if type(currentLeader) ~= "boolean" or currentLeader ~= true then return true end
				if BananaDKP:IsEventScheduled("DoLeaderSync") then return true end
				return currentLeader ~= STATUS_ACTIVE
			end,
			order = 105,
		},
		sitouts = {
			type = "group",
			name = L["Sitouts"],
			desc = L["List of players sitting out this raid who should still receive DKP. Click on a name to remove him from the list."],
			disabled = function() return type(currentLeader) ~= "boolean" or not currentLeader end,
			args = {
				add = {
					type = "text",
					name = L["Add player"],
					desc = L["Adds a player to the sitout list."],
					get = false,
					usage = "<player name>",
					set = function(v) BananaDKP:AddSitoutPlayer(v) end,
					order = 1,
				},
				clear = {
					type = "execute",
					name = L["Clear list"],
					desc = L["Clear the sitout list."],
					func = function() BananaDKP:ClearSitoutList() end,
					order = 2,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 3,
				},
			},
			order = 106,
			disabled = function()
				if not UnitInRaid("player") then return true end
				if not IsRaidLeader() and not IsRaidOfficer() then return true end
				if type(currentLeader) ~= "boolean" or currentLeader ~= true then return true end
				return false
			end,
		},
	},
}

-----------------------------------------------------------------------
-- Temporary Debug Logging
-----------------------------------------------------------------------

local function Log(msg)
	if not BananaDKP.debugFrame then return end
	BananaDKP.debugFrame:AddMessage(msg)
end

-----------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------

BananaDKP = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceConsole-2.0",
	"AceDB-2.0",
	"FuBarPlugin-2.0",
	"AceComm-2.0",
	"AceModuleCore-2.0",
	"AceHook-2.1"
)
local BananaDKP = BananaDKP

BananaDKP:SetModuleMixins("AceEvent-2.0")

-- Meh, can't set this in OnInitialize, since OnEmbedInitialize is called for
-- FBP before our OnInit, and FBP requires the property to be set then.
BananaDKP.hasIcon = "Interface\\Icons\\Ability_Hunter_AspectOfTheMonkey"

function BananaDKP:OnInitialize()
	local revision = tonumber((string.match("$Revision: 29748 $", "^.-(%d+).-$"))) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.BANANADKP_VERSION = self.version
	_G.BANANADKP_REVISION = self.revision

	self:RegisterDB("BananaDKPDB")
	self:RegisterDefaults("profile", {
		sitouts = {}, -- Remember this across sessions.
		dkpType = "Symphony", -- The DKP module to use.
		-- playerId = {}, -- Key: Name, Value: ID. Not using this, we're just
		-- using the player names for keys now - this should be fixed, or maybe
		-- an alias table could be added, for player alts that should have the
		-- same DKP.
	})

	self.OnMenuRequest = options

	self:RegisterChatCommand({ "/bananadkp" }, self.OnMenuRequest, "BANANADKP")

	self.debugFrame = ChatFrame1

	-- AceComm.enableLoopback = true
end

function BananaDKP:OnEnable(first)
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end

	if type(popup["BananaDKPLeaderPrompt"]) ~= "table" then
		popup["BananaDKPLeaderPrompt"] = {
			button1 = L["Accept"],
			button2 = L["Reject"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 20,
			showAlert = 1,
			OnAccept = function(data)
				acceptedLeader = data
			end,
			OnCancel = function(data, reason)
				if reason == "timeout" then
					acceptedLeader = data
				else
					acceptedLeader = nil
					BananaDKP:SendCommMessage("WHISPER", data, "RejectLeader")
				end
			end
		}
	end

	db = self.db.profile

	local currentModule = db.dkpType
	for name, module in self:IterateModules() do
		if name == currentModule then
			self:SetActiveDKPModule(name)
			break
		end
	end

	-- For hiding DKP whispers.
	self:Hook("ChatFrame_MessageEventHandler", "WhisperHandler", true)

	self:RegisterEvent("CHAT_MSG_COMBAT_HOSTILE_DEATH")
	self:RegisterEvent("CHAT_MSG_LOOT")
	self:RegisterBucketEvent("BananaDKP_Synchronize", 5)
	self:RegisterEvent("BananaDKP_JoinedGroup", "Toggle")
	self:RegisterEvent("BananaDKP_LeftGroup", "Toggle")
	self:RegisterEvent("CHAT_MSG_SYSTEM")

	currentStatus = STATUS_INACTIVE

	self:RebuildSitoutList()

	if (not first or GetNumRaidMembers() > 0) and not running then
		self:Toggle()
	end
end

-- Called when the user enables BananaDKP through the menu option.
function BananaDKP:Toggle()
	rejectAsLeader = 0
	recentWhispers = nil
	winner = nil
	currentLeader = nil

	if running then
		self:StopComm()
		currentStatus = STATUS_INACTIVE
		running = nil
	else
		if currentStatus ~= STATUS_CHECKING_NETWORK then
			self:StartComm()
			self:SendCommMessage("RAID", "LeaderPing")
			self:ScheduleEvent("LeaderPingTimeout", self.LeaderPingTimeout, 5, self)
			currentStatus = STATUS_CHECKING_LEADER
		end

		running = true
	end
	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- Events
-----------------------------------------------------------------------

function BananaDKP:CHAT_MSG_SYSTEM( msg )
	if msg:find("^"..ERR_RAID_YOU_LEFT) or msg:find(string.format(ERR_RAID_MEMBER_REMOVED_S, UnitName("player"))) then
		self:TriggerEvent("BananaDKP_LeftGroup")
	elseif msg:find(ERR_RAID_YOU_JOINED) then
		self:TriggerEvent("BananaDKP_JoinedGroup")
	end
end

-- Someone thinks we should be synchronizing.
function BananaDKP:BananaDKP_Synchronize()
	-- We get this event every time the DKP table changes, so clear our cache.
	for k, v in pairs(dkpValues) do
		dkpValues[k] = nil
	end

	if type(currentLeader) == "boolean" and currentLeader == true and currentStatus == STATUS_ACTIVE then
		self:SyncRaidToLeader()
	end
end

function BananaDKP:CHAT_MSG_LOOT(msg)
	if type(currentLeader) ~= "boolean" or currentLeader ~= true then return end

	local player, item = deformat(msg, LOOT_ITEM)
	if not player then
		item = deformat(msg, LOOT_ITEM_SELF)
		if item then
			player = UnitName("player")
		end
	end

	if type(item) == "string" and type(player) == "string" then
		local itemName, itemLink, itemRarity = GetItemInfo(item)
		-- Unconditionally notify modules regardless of itemRarity. Of course,
		-- most times modules will only care if itemRarity is higher than 3
		-- (epic or higher), but you never know.
		if type(itemName) == "string" and type(itemLink) == "string" and type(itemRarity) == "number" then
			for name, module in self:IterateModules() do
				if self:IsModuleActive(name) and type(module.ItemLooted) == "function" then
					module:ItemLooted(player, itemLink, itemRarity)
					break
				end
			end
		end
	end
end

function BananaDKP:CHAT_MSG_COMBAT_HOSTILE_DEATH(msg)
	if type(currentLeader) ~= "boolean" or currentLeader ~= true then return end
	local mob = deformat(msg, UNITDIESOTHER)
	if mob and BB:HasReverseTranslation(mob) then
		-- Since Babble-Boss contains a lot of mobs/adds that are not bosses,
		-- we scan the raid for people with this mob as target and check
		-- UnitClassification if it is worldboss.
		-- This means we don't get events for bosses in 5-man instances, though,
		-- even heroic.
		local found = nil
		if UnitExists("target") and UnitName("target") == mob and UnitClassification("target") == "worldboss" then
			found = true
		end
		for n, u in pairs(roster.roster) do
			local id = u.unitid .. "target"
			if UnitExists(id) and UnitName(id) == mob then
				if UnitClassification(id) == "worldboss" then
					found = true
				end
				break
			end
		end

		if found then
			self:TriggerEvent("BananaDKP_BossKilled", BB:GetReverseTranslation(mob))
		end
	end
end

-----------------------------------------------------------------------
-- Hooks
-----------------------------------------------------------------------

function BananaDKP:RegisterWhisperHandler(prefix, handlerFunc, prefixHelp)
	assert(type(prefix) == "string", "Invalid prefix.")
	prefix = prefix:lower()
	assert(not whisperHandlers[prefix], "Prefix already registered.")

	whisperHelp[prefix] = prefixHelp
	whisperHandlers[prefix] = handlerFunc
end

local function tablePrinter(t, recipient)
	for i, v in ipairs(t) do
		if type(v) == "string" then
			BananaDKP:Chat("BDKP: " .. v, "WHISPER", nil, recipient)
		elseif type(v) == "table" then
			tablePrinter(v, recipient)
		end
	end
end

function BananaDKP:WhisperHandler(event, ...)
	if (event == "CHAT_MSG_WHISPER" or event == "CHAT_MSG_WHISPER_INFORM") and type(arg1) == "string" and type(arg2) == "string" then
		local text = arg1:lower()
		if text:find("^b.-dkp") then
			if event == "CHAT_MSG_WHISPER_INFORM" then
				return
			end
			local player = arg2:lower()
			if type(recentWhispers) ~= "table" then
				recentWhispers = {}
			end
			local t = recentWhispers[player]
			if not t or (GetTime() - t > 1) then
				recentWhispers[player] = GetTime()
				local command, rest = select(3, text:find("^b.-dkp (%w+)(.*)"))
				if command then
					command = command:gsub("%A", ""):lower():trim()  -- So we don't disconnect if someone tries to send us item links.
				end
				if rest then
					rest = rest:gsub("%A", ""):lower():trim()
					if rest:len() < 3 then rest = nil end
				end
				if command and type(whisperHandlers[command]) == "function" then
					local ret = whisperHandlers[command](event, rest or arg2:lower())
					if type(ret) == "string" and ret:trim():len() > 3 then
						self:Chat(string.format(L["BananaDKP: %s."], ret), "WHISPER", nil, arg2)
					elseif type(ret) == "table" then
						tablePrinter(ret, arg2)
					end
				elseif command and command == "help" then
					local ret = self:GetHelpText(rest)
					if type(ret) == "string" then
						self:Chat(ret, "WHISPER", nil, arg2)
					elseif type(ret) == "table" then
						for i, v in ipairs(ret) do
							self:Chat(v, "WHISPER", nil, arg2)
						end
					end
				else
					player = command or arg2
					local dkp = dkpValues[player]
					if type(dkp) == "number" then
						self:Chat(string.format(L["BananaDKP: DKP for %s: %s."], player, tostring(dkp)), "WHISPER", nil, arg2)
					else
						self:Chat(L["BananaDKP: Unknown command or player not found."], "WHISPER", nil, arg2)
					end
				end
				return
			end
		elseif currentStatus == STATUS_TAKING_BIDS and text:find("^bid") then
			self:NewBidder(arg2)
		end
	elseif event == "CHAT_MSG_RAID" or event == "CHAT_MSG_RAID_LEADER" and type(arg2) == "string" and type(currentLeader) == "boolean" and currentLeader == true then
		if arg2 == UnitName("player") and arg1:lower():find(L["^bid:"]) then
			if currentStatus ~= STATUS_TAKING_BIDS then
				currentStatus = STATUS_TAKING_BIDS
				self:ScheduleEvent("BidTimeout", self.BidTimeout, 15, self)
			else
				-- We're already taking bids, but someone won it and we're going
				-- again with a new item.
				winner = nil
				self:ScheduleEvent("BidTimeout", self.BidTimeout, 15, self)
			end
		elseif currentStatus == STATUS_TAKING_BIDS and arg1:lower():find(L["^bid"]) then
			self:NewBidder(arg2)
		end
	end
	return self.hooks["ChatFrame_MessageEventHandler"](event, ...)
end

function BananaDKP:NewBidder(bidder)
	local dkp = dkpValues[arg2]
	if not winner or winner == bidder then
		winner = bidder
	else
		local winnerDkp
		if type(winner) == "string" then
			winnerDkp = dkpValues[winner]
		elseif type(winner) == "table" then
			winnerDkp = dkpValues[winner[1]]
		end
		if not winnerDkp then winnerDkp = -100 end
		if dkp > winnerDkp then
			winner = bidder
		elseif dkp == winnerDkp then
			if type(winner) ~= "table" then
				winner = {winner}
			end
			local skip = nil
			for i, v in ipairs(winner) do
				if v == bidder then
					skip = true
				end
			end
			if not skip then
				table.insert(winner, bidder)
			end
		end
	end
	self:ScheduleEvent("BidTimeout", self.BidTimeout, 15, self)
end

function BananaDKP:GetHelpText(topic)
	if topic then
		if whisperHelp[topic] then
			return whisperHelp[topic]
		elseif topic == "bid" then
			return L["Chat-Help-Bid"]
		elseif topic == "dkp" then
			return L["Chat-Help-Dkp"]
		end
	else
		local help = {}
		table.insert(help, L["Chat-Help-Dkp"])
		table.insert(help, L["Chat-Help-Bid"])
		for k, v in pairs(whisperHelp) do
			table.insert(help, v)
		end
		return help
	end
end

-----------------------------------------------------------------------
-- Communication Handling
-----------------------------------------------------------------------

--
-- Communication flow description (horribly outdated)
--
-- Note that the communication for BananaDKP is not secure, it can be modified
-- by local clients as they see fit.
-- This means that any DKP value that anyone has should be subject to public
-- inquiery at all times.
--
-- The communication could be made secure, if that is desired, by using any of
-- the encryption libraries available at the wowace SVN.
--
-- 1. You form a new raid or join an existing one, provided you have set the
--    DKP ID, go to :Init.
--
-- :Init
-- 1. You send a global VersionPing, and wait 10 seconds for responses. If
--    you don't get any, go to :Done.
-- 2. If you got ANY response with a version that does not match your own, go
--    to :Done.
-- 3. If you got less than 3 responses, go to :Done.
-- 4. Go to :LeaderPing.
--
-- :LeaderPing
-- 1. You send a LeaderPing broadcast. If you get no leader response in 5
--    seconds, go to :NoLeader.
-- 2. Send a DKPSyncRequest to the leader. If you don't get any response
--    in 10 seconds, set our state to failed and enable the "Get current
--    leader" option. If you click that, go to :Init.
-- 3. Go to :Done.
-- 
-- :NoLeader
-- 1. If you are the raid leader or officer, the option "Set myself as leader"
--    is now available. If you click that, go to :SetSelfLeader.
-- 2. Go to :Done.
--
-- :SetSelfLeader
-- 1. You send a SetLeader broadcast, and start a timeout of 30 seconds where
--    people can either accept or reject you as the leader.
--    If you get 1 or more rejects OR no accepts, you are not set as the
--    leader, and go to :Done.
--    If you get no rejects AND an accept, you are set as the leader.
-- 2. Send a broadcasted DKPSyncRequest, where people respond to you in whisper
--    with their DKP data, you sync it with you local table, and then send a
--    DKPLeaderSync with your data, so people can update their local DKP tables
--    with the global sync.
-- 3. Go to :Done.
--
-- :Done
--

local prefixRegistered = nil
function BananaDKP:StartComm()
	if not prefixRegistered then
		self:SetCommPrefix("BananaDKP")
		self:SetDefaultCommPriority("BULK")
		self:RegisterMemoizations({
			"LeaderPing",
			"LeaderPong",
			"SetLeader",
			"LeaderSet",
			"LeaderDispute",
			"RejectLeader",
			"DKPSyncRequest",
			"DKPSyncResponse",
			"DKPLeaderSync",
		})
		prefixRegistered = true
	end

	self:RegisterComm("BananaDKP", "RAID", "OnDistributedComm")
	self:RegisterComm("BananaDKP", "WHISPER", "OnIndividualComm")
end

function BananaDKP:StopComm(reason)
	self:UnregisterAllComms()
	self:CancelAllScheduledEvents()

	currentStatus = reason or 0

	self:UpdateDisplay()
end

function BananaDKP:OnDistributedComm(prefix, sender, distribution, command, ...)
	if distribution ~= "RAID" then return end
	if type(command) == "string" and commHandlers[command] then
		Log("Recieved RAID ["..command.."] from ["..sender.."].")
		commHandlers[command](self, sender, ...)
	end
end

function BananaDKP:OnIndividualComm(prefix, sender, distribution, command, ...)
	if distribution ~= "WHISPER" then return end
	if type(command) == "string" and commHandlers[command] then
		Log("Recieved WHISPER ["..command.."] from ["..sender.."].")
		commHandlers[command](self, sender, ...)
	end
end

--
-- Received broadcasted when someone is wondering who the leader of
-- the current DKP stream is.
--
-- Everyone EXCEPT the leader will reply to this.
--
commHandlers["LeaderPing"] = function(self, sender)
	-- currentLeader == true means we are the leader, so we don't respond.
	if not currentLeader then return end
	Log("  Replying with ["..tostring(currentLeader).."].")
	if type(currentLeader) == "boolean" then
		self:SendCommMessage("WHISPER", sender, "LeaderPong", UnitName("player"))
	else
		self:SendCommMessage("WHISPER", sender, "LeaderPong", currentLeader)
	end
end

--
-- Recieved whisper when someone has responded to your LeaderPing request.
--
local lastLeaderPong = nil
commHandlers["LeaderPong"] = function(self, sender, leader)
	assert(leader, string.format("Got a invalid leader pong from %q.", tostring(sender)))
	local leaderUnit = roster:GetUnitFromName(leader)
	assert(leaderUnit, string.format("Leader %q from %q does not exist in the current roster.", tostring(leader), tostring(sender)))

	if type(lastLeaderPong) == "string" and lastLeaderPong == leader then return end

	self:CancelScheduledEvent("LeaderPingTimeout")

	-- If we get another leader pong, and it's not the same as last time, we
	-- unregister all comms and set our status to failed.
	if type(lastLeader) == "string" and lastLeader ~= leader then
		self:StopComm(STATUS_LEADER_DISPUTE)
		return
	end

	Log("Setting leader to ["..leader.."] and requesting DKP sync.")

	lastLeaderPong = leader

	currentStatus = STATUS_DKP_SYNC

	-- Now we need to sync ourselves with the current leader.
	self:SendCommMessage("WHISPER", leader, "DKPSyncRequest")
	self:ScheduleEvent("DKPSyncRequestTimeout", self.DKPSyncRequestTimeout, 10, self)

	self:UpdateDisplay()
end

--
-- Recieved broadcasted when someone wants to set themselves as the leader of
-- the current DKP stream.
--
-- Everyone should reply with accept or deny.
--
commHandlers["SetLeader"] = function(self, sender)
	if type(currentLeader) == "string" or type(currentLeader) == "boolean" then
		error("There seems to already be a DKP stream leader.")
	end

	Log("Got a SetLeader request from "..sender..".")

	StaticPopupDialogs["BananaDKPLeaderPrompt"].text = string.format(L["%s wants to be the DKP stream leader for this raid.\n\nDo you accept? (You will auto-accept in 20 seconds.)"], sender)
	local popup = StaticPopup_Show("BananaDKPLeaderPrompt")
	if popup then popup.data = sender end
end

commHandlers["LeaderDispute"] = function(self, sender)
	-- Should notify with a chat message that sender has disputed the current
	-- leader.
	Log("Leader dispute!")
	currentStatus = STATUS_LEADER_DISPUTE
	self:CancelScheduledEvent("LeaderSetTimeout")
	self:UpdateDisplay()
end

commHandlers["LeaderSet"] = function(self, sender)
	-- If sender is not the leader we accepted, then send a LeaderDispute.
	if acceptedLeader ~= sender then
		self:SendCommMessage("RAID", "LeaderDispute")
		acceptedLeader = nil
		return
	end
	acceptedLeader = nil
	currentLeader = sender
	Log("Leader set to "..sender..".")
	currentStatus = STATUS_ACTIVE
	self:UpdateDisplay()
end

--
-- Recieved whisper when someone has rejected you as the current DKP stream
-- leader.
--
commHandlers["RejectLeader"] = function(self, sender)
	assert(self:IsEventScheduled("SetLeaderTimeout"), "We got a RejectLeader, but we're not trying to become one.")
	Log("Rejected Leader + 1 from " .. tostring(sender) .. ".")
	rejectAsLeader = rejectAsLeader + 1
end

--
-- Recieved whisper or global when someone wants to sync their dkp table with ours.
--
commHandlers["DKPSyncRequest"] = function(self, sender)
	if sender ~= currentLeader or (type(currentLeader) == "boolean" and currentLeader ~= true) then
		error("Got a DKP sync request from "..sender..", but he is not the current DKP stream leader.")
	end
	Log("DKP Sync Request from "..sender..".")
	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			self:SendCommMessage("WHISPER", sender, "DKPSyncResponse", module:GetDKPSynchronizationData(sender))
			break
		end
	end
end

--
-- Recieved whisper when someone has sent us a sync'ed dkp table.
--
commHandlers["DKPSyncResponse"] = function(self, sender, ...)
	if type(currentLeader) ~= "boolean" or currentLeader ~= true then
		error("We got a DKP sync from "..sender..", but we're not the current DKP stream leader.")
	end
	Log("DKP Sync Response from "..sender..".")

	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			module:SynchronizeDKP(sender, ...)
			break
		end
	end

	self:ScheduleEvent("UpdateDisplay", self.UpdateDisplay, 5, self)
end

--
-- Recieved global when the leader is done sync'ing with the entire raid, so
-- that everyone gets up to speed.
--
commHandlers["DKPLeaderSync"] = function(self, sender, ...)
	if type(currentLeader) ~= "string" or sender ~= currentLeader then
		error("Got a DKPLeaderSync from "..sender..", but he's not the current raid leader.")
	end
	Log("DKP Leader Sync from "..sender..".")
	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			module:SynchronizeDKP(sender, ...)
			break
		end
	end

	self:ScheduleEvent("UpdateDisplay", self.UpdateDisplay, 5, self)
end

function BananaDKP:Chat(text, chatType, language, destination)
	if ctl then
		ctl:SendChatMessage("BULK", "BananaDKP", text, chatType, language, destination)
	else
		SendChatMessage(text, chatType, language, destination)
	end
end

-----------------------------------------------------------------------
-- Communication Timeout Handling
-----------------------------------------------------------------------

function BananaDKP:BidTimeout()
	if type(winner) == "table" then
		self:Chat(string.format(L["Tie: %s."], table.concat(winner, ", ")), "RAID")
	elseif type(winner) == "string" then
		self:Chat(string.format(L["%s wins!"], tostring(winner)), "RAID")
	end

	winner = nil
	currentStatus = STATUS_ACTIVE
end

-- Timeout is always triggered for SetLeader.
function BananaDKP:SetLeaderTimeout()
	if rejectAsLeader == 0 then
		-- We are accepted as the leader.
		self:SendCommMessage("RAID", "LeaderSet")
		currentLeader = true
		currentStatus = STATUS_ACTIVE
		-- Sync all the DKP in 5 seconds.
		self:ScheduleEvent("GlobalDKPSync", self.SyncRaidDKP, 5, self)
		Log("We are set as leader; Requesting global DKP sync in 5 seconds ...")
	else
		currentStatus = STATUS_LEADER_DISPUTE
		Log("Leader dispute, disabling.")
		self:Toggle()
	end
	rejectAsLeader = 0
	self:UpdateDisplay()
end

function BananaDKP:LeaderPingTimeout()
	currentStatus = STATUS_NO_LEADER
	Log("Leader ping timeout. You can now set yourself as the DKP stream leader.")
	self:UpdateDisplay()
end

function BananaDKP:DKPSyncRequestTimeout()
	currentStatus = STATUS_NO_DKP_FROM_LEADER
	Log("DKP sync request timeout.")
	self:Toggle()
end

-----------------------------------------------------------------------
-- DKP Sync
-----------------------------------------------------------------------

function BananaDKP:SyncRaidDKP()
	if type(currentLeader) ~= "boolean" or currentLeader ~= true then
		error("You can't sync the raid DKP if you're not the DKP stream leader.")
	end

	currentStatus = STATUS_DKP_SYNC

	Log("Requesting raid DKP sync ...")
	self:SendCommMessage("RAID", "DKPSyncRequest")

	-- Give people 30 seconds to sync before we transmit our data.
	self:ScheduleEvent("DoLeaderSync", self.SyncRaidToLeader, 30, self)

	self:UpdateDisplay()
end

function BananaDKP:SyncRaidToLeader()
	if type(currentLeader) ~= "boolean" or currentLeader ~= true then
		error("You can't sync the raid DKP if you're not the DKP stream leader.")
	end

	currentStatus = STATUS_DKP_SYNC

	Log("Synchronizing the raid with our DKP database ...")

	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			self:SendCommMessage("RAID", "DKPLeaderSync", module:GetDKPSynchronizationData("RAID"))
			break
		end
	end

	self:ScheduleEvent("DoLeaderSync", function()
		currentStatus = STATUS_ACTIVE
		self:UpdateDisplay()
	end, 10, self)

	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- Command Handling
-----------------------------------------------------------------------

function BananaDKP:SetSelfAsLeader()
	assert(UnitInRaid("player"), "You are not in a raid group.")
	assert(IsRaidLeader() or IsRaidOfficer(), "You are not a raid officer or leader.")
	assert(currentStatus == STATUS_NO_LEADER, "We have not yet determined whether the raid has a current stream leader or not.")

	Log("Setting myself as DKP stream leader in 30 seconds ...")
	self:SendCommMessage("RAID", "SetLeader")
	self:ScheduleEvent("SetLeaderTimeout", self.SetLeaderTimeout, 30, self)
	currentStatus = STATUS_SETTING_LEADER

	self:UpdateDisplay()
end

-----------------------------------------------------------------------
-- FuBar Handling
-----------------------------------------------------------------------

function BananaDKP:GetStatusColor(status)
	if status < 100 then
		return colors["success"]
	elseif status < 200 then
		return colors["inactive"]
	elseif status < 300 then
		return colors["progress"]
	end
	return colors["failure"]
end

local currentFuBarText = nil
function BananaDKP:OnTextUpdate()
	local color = self:GetStatusColor(currentStatus)
	currentFuBarText = color .. (L[currentStatus] or L["Unknown status!"]) .. "|r"
	self:SetText(currentFuBarText)
end

function BananaDKP:OnTooltipUpdate()
	local cat = tablet:AddCategory(
		"columns", 2,
		"child_textR", 1,
		"child_textG", 1,
		"child_textB", 0,
		"child_text2R", 1,
		"child_text2G", 1,
		"child_text2B", 1
	)
	cat:AddLine("text", L["Status"], "text2", currentFuBarText)
	if type(currentLeader) == "string" then
		cat:AddLine("text", L["DKP Leader"], "text2", coloredName[currentLeader])
	end
	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			cat:AddLine("text", L["Type"], "text2", name)
			break
		end
	end
	cat = tablet:AddCategory(
		"columns", 2,
		"text", L["Player"],
		"text2", L["DKP"],
		"child_textR", 1,
		"child_textG", 1,
		"child_textB", 0,
		"child_text2R", 1,
		"child_text2G", 1,
		"child_text2B", 1
	)
	for n, u in pairs(roster.roster) do
		if u.class ~= "PET" then
			cat:AddLine("text", coloredName[n], "text2", dkpValues[n] or 0)
		end
	end
end

-----------------------------------------------------------------------
-- DKP Module Handling
-----------------------------------------------------------------------

function BananaDKP:GetCurrentLeader()
	return currentLeader
end

--[[ Unused
function BananaDKP:GetPlayerID(name)
	assert(type(name) == "string", "Player names can only be strings.")
	name = name:lower()
	local db = self.db.profile.playerId
	if not db[name] then
		local nextId = 0
		for k, v in pairs(db) do
			if v > nextId then nextId = v end
		end
		nextId = nextId + 1
		db[name] = nextId
	end
	return db[name]
end

function BananaDKP:GetPlayerName(id)
	return playerName[id]
end
]]
function BananaDKP:GetDKP(player)
	local dkp = nil
	for name, module in self:IterateModules() do
		if self:IsModuleActive(name) then
			dkp = module:GetDKP(player)
			break
		end
	end
	return dkp
end

function BananaDKP:SetActiveDKPModule(moduleName)
	assert(self:HasModule(moduleName), moduleName .. " doesn't seem to exist.")
	assert(not running, "Can't change the DKP type while active.")

	db.dkpType = moduleName

	-- Clear the registered whisper handlers.
	whisperHandlers = {}
	whisperHelp = {}

	for name, module in self:IterateModules() do
		if name == moduleName then
			self:ToggleModuleActive(module, true)
			options.args.dkpTypeOptions.args = module.options
		else
			self:ToggleModuleActive(module, false)
		end
	end

	self:UpdateDisplay()
end

function BananaDKP:RegisterDKPType(name, module)
	if module:IsRegistered() then
		error(name .. " has already been registered.")
	end
	if type(module.defaultDatabase) ~= "table" then
		error(name .. " needs a default database.")
	end
	if type(module.defaultDatabase.databaseVersion) ~= "number" then
		error(name .. "'s database needs a databaseVersion number field.")
	end
	if type(module.GetDKP) ~= "function" then
		error(name .. " needs a GetDKP method.")
	end
	if type(module.GetLootedItems) ~= "function" then
		error(name .. " needs a GetLootedItems method.")
	end
	if type(module.SynchronizeDKP) ~= "function" then
		error(name .. " needs a SynchronizeDKP method.")
	end
	if type(module.GetDKPSynchronizationData) ~= "function" then
		error(name .. " needs a GetDKPSynchronizationData method.")
	end

	module.db = self:AcquireDBNamespace(name)
	self:RegisterDefaults(name, "profile", module.defaultDatabase)

	self:ToggleModuleActive(module, false)

	local opt = options.args.dkpType.args

	opt[name] = {
		type = "toggle",
		name = name,
		desc = string.format(L["Activate %s."], name),
		get = function() return BananaDKP.db.profile.dkpType == name end,
		set = function()
			BananaDKP:SetActiveDKPModule(name)
		end,
		disabled = function() return running end,
	}

	if type(module.OnRegister) == "function" then
		module:OnRegister()
	end
end

-----------------------------------------------------------------------
-- DKP Type prototype
-----------------------------------------------------------------------

BananaDKP.modulePrototype.core = BananaDKP

function BananaDKP.modulePrototype:OnInitialize()
	assert(not self.registered, self.name .. " has already been registered.")
	self.core:RegisterDKPType(self.name, self)
end

function BananaDKP.modulePrototype:IsRegistered()
	return self.registered
end

-----------------------------------------------------------------------
-- Sitout list handling
-----------------------------------------------------------------------

function BananaDKP:ClearSitoutList()
	db.sitouts = {}
	self:RebuildSitoutList()
end

function BananaDKP:AddSitoutPlayer(name)
	db.sitouts[name] = true
	self:RebuildSitoutList()
end

function BananaDKP:RemoveSitoutPlayer(name)
	db.sitouts[name] = nil
	self:RebuildSitoutList()
end

function BananaDKP:RebuildSitoutList()
	local opts = options.args.sitouts.args

	for k, v in pairs(opts) do
		if k ~= "add" and k ~= "clear" and k ~= "spacer" then
			opts[k] = nil
		end
	end

	for k, v in pairs(db.sitouts) do
		opts[k] = {
			type = "execute",
			name = k,
			desc = string.format(L["Remove %s."], k),
			func = function() BananaDKP:RemoveSitoutPlayer(k) end,
			order = 100,
		}
	end
end

